# ---
# title: "Majority voting function : R script for land cover classification and mapping of a polar desert in the Canadian Arctic Archipelago"
# date: "21 February 2023"
# ---


################################ Libraries ################################
library(terra)
library(raster)


################################ Majority voting function ################################
#### Description ####
# Method to create a SpatRaster object that represent the values of the majority voting procedure from a multiple layers SpatRaster. A cut option is included to enable users to run the majority voting procedure with low random access memory (RAM). The procedure can take a longer time in that case.

#### Arguments ####
# data              A SpatRaster object containing multiple layers.

# resolution        The resolution of the input data. The default is 0.5. 

# cut               If TRUE, will separate the input data in a number of smaller SpatRasters. The input SpatRaster will be cut along the Y axis of the data.

# nslice            If cut = TRUE, determine the number of smaller SpatRaster (slices) to be performed. The number of pixels on the Y axis divided by nslice must result in a whole number.

# priority          In the event of a tie in the majority voting for a given pixel, its value will be set accordingly to the value of the layer that is present in the tie and have the highest priority number. The default set the highest priority to the first layer and so on. If the default is changed, a numeric vector of equal length to the number of layers must be provided. The first number of the vector must be the number of layer of first priority, the second number, the number of the layer of the second most priority and so on.

# path.slice        If cut = TRUE, a directory path to an empty folder. The slices will be exported in that folder.

# path              A directory path to an empty folder. The output SpatRaster will be exported in that folder.

# filename          The name of the output SpatRaster with the file extension.

Maj.voting.fct<-function(data,resolution = 0.5, cut = TRUE, nslice,priority=1:nlyr(data),path.slice,path,filename){
  
  require(terra)
  
  # Collecting information from the input data
  
  data.nlayers<-nlyr(data)
  
  data.ncol<-ncol(data)
  
  data.nrow<-nrow(data)
  
  data.ymin<-ymin(data)
  
  data.ymax<-ymax(data)
  
  data.xmin<-xmin(data)
  
  data.xmax<-xmax(data)
  
  df.list<-list()
  
  if(cut == TRUE){
    
    # Separation of the input data in a nslice of smaller SpatRasters
    
    cat("","\n")
    cat("- Initializing slicing procedure -","\n")
    cat("","\n")
    
    pb <- txtProgressBar(min = 0,max = nslice, style = 3, width = 50,char = "=") # Initializing the progression bar for the slicing process
    
    # Determination of the Y axis size of the slices (y.frac)
    
    y.diff<-data.ymax-data.ymin
    y.frac<-y.diff/nslice
    
    # Loop to create slices
    
    for(a in 1:nslice){
      
      setTxtProgressBar(pb, a) # Actualization of the progression bar
      
      stack.sliced<-rast() # Creation of an empty SpatRaster that will include all the small SpatRasters of a given portion of the input data from every layers
      
      # Determination of the portion of the input data to consider (the slice a)
      
      new.ymin<-data.ymin+(y.frac*(a-1))
      new.ymax<-new.ymin+y.frac
      
      # Loop to fill the SpatRaster (stack.sliced) with the small SpatRasters of the slice a
      
      for(b in 1:data.nlayers){
        
        new<-rast(xmin=data.xmin, xmax=data.xmax, ymin=new.ymin, ymax=new.ymax)
        res(new)<-resolution
        interval<-((ncell(new)*(a-1))+1):(ncell(new)*a)
        crs(new)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
        values(new)<-values(data[[b]])[interval]
        suppressWarnings(stack.sliced<-c(stack.sliced,new)) # Adding the new sliced portion of the layer b to the SpatRaster
        
      }
      
      
      new<-as.data.frame(extract(stack.sliced,1:ncell(stack.sliced))) # Creation of a data.frame containing the values of a slice (rows) for each layers (columns)
      df.list[[a]]<-new # Adding the data.frame of the slice a to the list
      
    }
    
    close(pb) # Closing the progression bar
    
    
  }else{
    
    df.list[[1]]<-as.data.frame(extract(data,1:ncell(data))) # Creation of a data.frame containing the values (rows) of the input data for each layers (columns)
    
  }
  
  # Majority voting 
  
  cat("","\n")
  cat("- Initializing majority voting procedure -","\n")
  cat("","\n")
  
  pb <- txtProgressBar(min = 0,max = length(df.list), style = 3, width = 50,char = "=") # Initializing the progression bar for the majority voting process
  
  for(d in 1:length(df.list)){
    
    n.lines<-length(df.list[[1]][[1]]) # Finding the number of rows of all data.frames in the list
    
    Maj.voting<-c() # Creating an empty vector that will contains the values of the majority voting [for this slice (if cut = TRUE)]
    
    for(e in 1:n.lines){
      
      indic<-"NOT NA" # Setting a default value of the indicator for the missing data verification
      
      # Missing data verification
      
      # Loop for each column in the data.frame d
      
      for(i in 1:length(df.list[[d]])){
        
        if(is.na(df.list[[d]][e,i])==FALSE){
          
          break
          
        } # For the line e in the data.frame d, is the value in the column i is different than NA? if yes, then there is at least one value for this line, so the verification can stop (break)
        if(i==length(df.list[[d]])){
          
          most_freq_cat<-NA
          
          indic<-"NA"
        } # If all the values from all the columns were NA, than set the most frequent value for that line to NA and change the indicator to "NA"
      }
      
      # Majority voting procedure for non NA values
      
      if(indic == "NOT NA"){
        
        freqs<-as.data.frame(table(t(df.list[[d]][e,]))) # Creating a data.frame of the frequencies of values in the line e of the data.frame d
        colnames(freqs)[1] = "Category" 
        freqs$Category<-as.character(freqs$Category)
        
        
        if(length(freqs$Category)>1){
          
          maximum<-freqs[which(freqs$Freq == max(freqs$Freq)),] # Finding the category associated with the maximum frequency in that line
          
          if(length(maximum$Category)>1){
            
            # The case of tie between different maximum frequencies
            
            # Loop for each layer
            
            for(f in 1:length(df.list[[d]][1,])){
              
              # Checking if the layer with the priority f is absent of the maximum tie
              
              if(length(maximum$Category[which(maximum$Category == df.list[[d]][e,priority[f]])])==0){
                
              }else{
                
                most_freq_cat<-as.numeric(maximum$Category[which(maximum$Category == df.list[[d]][e,priority[f]])])
                
                break
                
              }# if the layer with the priority f is present, than that value becomes the most frequent category
            }
            
          }else{
            
            most_freq_cat<-as.numeric(maximum$Category)
            
          }# If there is only one maximum frequency, than that value becomes the most frequent category
          
        }else{
          
          most_freq_cat<-as.numeric(freqs$Category)
          
        }# If there is only one category, than that value becomes the most frequent category
      }
      
      
      Maj.voting<-c(Maj.voting,most_freq_cat) # Adding the majority voting value to the vector
      
    }
    
    # Exporting the majority voting output
    
    if(cut == TRUE){
      
      # Creating and exporting the sliced output SpatRaster a
      
      ymin.new<-data.ymin+(y.frac*(d-1))
      ymax.new<-ymin.new+y.frac
      ncol.new<-data.ncol/nslice
      
      new<-rast(ncols=ncol.new,nrows=data.nrow,xmin=data.xmin,xmax=data.xmax,ymin=ymin.new,ymax=ymax.new)
      res(new)<-resolution
      crs(new)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
      values(new)<-Maj.voting
      names(new)<-"Maj_voting"
      
      nrast<-paste("Raster_no",(length(df.list)-(d-1)),".tif",sep="")
      setwd(path.slice)
      writeRaster(new, file=nrast,overwrite=TRUE)
      
    }else{
      
      new<-rast(ncols=data.ncol,nrows=data.nrow,xmin=data.xmin,xmax=data.xmax,ymin=data.ymin,ymax=data.ymax)
      res(new)<-resolution
      crs(new)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
      values(new)<-Maj.voting
      names(new)<-"Maj_voting"
      
      setwd(path)
      writeRaster(new, file=filename,overwrite=TRUE) # Exporting the SpatRaster
      maj.vot.raster <<- new # Exporting the SpatRaster to the R environment
    } # If cut = FALSE, than export the output SpatRaster directly
    
    
    
    setTxtProgressBar(pb, d) # Actualization of the progression bar
    
  }
  close(pb)  # Closing the progression bar
  
  
  if(cut == TRUE){
    # Importing back all the output slices 
    
    raslist <- list.files(path.slice, pattern='.tif$', all.files=TRUE, full.names=TRUE)
    
    setwd(path)
    
    require(raster)
    
    ras.list<-list()
    
    for(g in 1:length(raslist)){
      
      ras.list[[g]]<-raster(raslist[[g]])
      
    }
    
    ras.list$filename <- filename # Associating the list with the filename
    maj.vot.raster <- do.call(merge, ras.list) # Merging all the output slices into one output RasterLayer
    maj.vot.raster <<- rast(maj.vot.raster) # Exporting the output SpatRaster in the R environment
  }
  
  
}


#### Example 1 ####
# Creating three SpatRasters of the same size
a<-rast(ncols=100,nrows=50)
crs(a)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
values(a)<-c(sample(c(1:7),(ncell(a)-100),replace=TRUE),rep(NA,100))

b<-rast(ncols=100,nrows=50)
crs(b)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
values(b)<-c(sample(c(1:7),(ncell(b)-30),replace=TRUE),rep(NA,30))

c<-rast(ncols=100,nrows=50)
crs(c)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
values(c)<-c(sample(c(1:7),(ncell(c)-50),replace=TRUE),rep(NA,50))

# Combining the three SpatRasters
ras1<-c(a,b,c)

# Using the function
Maj.voting.fct(data=ras1,resolution=c(3.6,3.6),cut=T,nslice=10,path.slice = "C:\\Users\\dese0042\\Desktop\\Maj_voting\\example1\\slices",path = "C:\\Users\\dese0042\\Desktop\\Maj_voting\\example1\\result_map",filename= "example1.tif")

# Plotting the output map
plot(maj.vot.raster)


#### Example 2 ####
# Creating four SpatRasters of the same size
d<-rast(ncols=5,nrows=5)
crs(d)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
values(d)<-c(sample(c(1:7),(ncell(d)-3),replace=TRUE),rep(NA,3))

e<-rast(ncols=5,nrows=5)
crs(e)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
values(e)<-c(sample(c(1:7),(ncell(e)-3),replace=TRUE),rep(NA,3))

f<-rast(ncols=5,nrows=5)
crs(f)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
values(f)<-c(sample(c(1:7),(ncell(f)-3),replace=TRUE),rep(NA,3))

g<-rast(ncols=5,nrows=5)
crs(g)<-"+proj=utm +zone=20 +datum=NAD83 +units=m +no_defs"
values(g)<-c(sample(c(1:7),(ncell(g)-3),replace=TRUE),rep(NA,3))

# Combining the three SpatRasters
ras2<-c(d,e,f,g)

# Using the function
Maj.voting.fct(data=ras2,resolution=c(72,36),cut=F,path = "C:\\Users\\dese0042\\Desktop\\Maj_voting\\example2",filename= "example2.tif")

# Plotting the output map
plot(maj.vot.raster)


#### Data from the article ####
setwd("C:\\Users\\dese0042\\Desktop\\maps\\")
raslist <- list.files(getwd(), pattern='.tif$', all.files=TRUE, full.names=TRUE) # Will collect all tif images from the filepath
rasvars<- rast(raslist) # Create a SpatRaster with multiple layers
rasvars

# Using the function
Maj.voting.fct(data=rasvars,resolution=c(0.5,0.5),cut=T,nslice=2687,path.slice = "C:\\Users\\dese0042\\Desktop\\Maj_voting\\Majority_voting\\slices",path = "C:\\Users\\dese0042\\Desktop\\Maj_voting\\Majority_voting\\result_map",filename= "ensemble_classifier.tif")

# Plotting the output map
plot(maj.vot.raster)
