# ---
# title: "Predictor selection, classification and validation : R script for land cover classification and mapping of a polar desert in the Canadian Arctic Archipelago"
# date: "21 February 2023"
# ---


################################ Libraries ################################
library(randomForest)
library(psych)
library(sp)
library(rgdal)
library(terra)
library(e1071)
library(raster)
library(lattice)
library(ggplot2)
library(caret)
library(Boruta)
library(corrplot)


################################ Import data ################################
#### Import rasters ####
setwd("C:\\Users\\dese0042\\Desktop\\layers\\")
raslist <- list.files(getwd(), pattern='.tif$', all.files=TRUE, full.names=TRUE)
rasters<- stack(raslist)


#### Import ground references ####
reference<- shapefile("C:\\Users\\dese0042\\Desktop\\reference\\reference.shp")
reference$Class <- as.factor(reference$Class)


#### Extract mean value of raster layers (predictors) within each buffer ####
vec<-c()
for(i in 1:length(reference$Class)){
  if(reference$Class[i]=="snow" || reference$Class[i]=="water"){
    vec<-c(vec,2)
  }else{
    vec<-c(vec,6)
  }
}
vec

ref_extract<-as.data.frame(extract(rasters, reference, buffer=vec, fun='mean', na.rm=TRUE))
str(ref_extract)

reference@data=data.frame(reference@data, ref_extract[match(rownames(reference@data), rownames(ref_extract)),])
str(reference@data)
reference@data<- na.omit(reference@data)
complete.cases(reference@data)

dataframe_ref<- as.data.frame(reference@data)


#### Mean and standard deviation ####
mean_table <-aggregate(dataframe_ref[, 2:39], list(dataframe_ref$Class), mean)
write.table(mean_table,paste("C:\\Users\\dese0042\\Desktop\\Mean",".txt",sep=""),col.names = T,row.names = T,sep = "\t")

sd_table <-aggregate(dataframe_ref[, 2:39], list(dataframe_ref$Class), sd)
write.table(sd_table,paste("C:\\Users\\dese0042\\Desktop\\Standard_deviation",".txt",sep=""),col.names = T,row.names = T,sep = "\t")


#### Divide ground references in training dataset (80%) and validation dataset (20%) ####
hre_seed<- 50
set.seed(hre_seed)

intraining <- createDataPartition(reference@data$Class, p = .80, list = FALSE)

training <- reference@data[intraining,]
validation <- reference@data[-intraining,]

dim(training)
dim(validation)


################################ Predictor selection ################################
#### Area Under the receiver operator characteristic Curve (AUC) ####
d2<-dataframe_ref[,-1]
d<-dataframe_ref[,1]

ROC<-filterVarImp(d2,d, nonpara = TRUE)
ROC

write.table(ROC,paste("C:\\Users\\dese0042\\Desktop\\AUC",".txt",sep=""),col.names = T,row.names = T,sep = "\t")


#### Boruta ####
boruta_output <- Boruta(Class ~., data=na.omit(dataframe_ref), doTrace=0) 
boruta_signif <- names(boruta_output$finalDecision[boruta_output$finalDecision %in% c("Confirmed", "Tentative")])
print(boruta_signif)
plot <-plot(boruta_output, cex.axis=.6, las=2, xlab="", main="Variable Importance")  
plot

getSelectedAttributes(boruta_output, withTentative = F)
bot2<-attStats(boruta_output) 
print(bot2)

write.table(bot2,paste("C:\\Users\\dese0042\\Desktop\\Botura",".txt",sep=""),col.names = T,row.names = T,sep = "\t")


#### Correlation matrix ####
correlation_matrix <- cor(d2)
print(correlation_matrix)

highly_correlated <- findCorrelation(correlation_matrix, cutoff=1)
print(highly_correlated)

corrplot(cor(correlation_matrix), method="color", type = "lower", tl.cex=0.5, tl.col="black")
corrplot(cor(correlation_matrix), method="number", type = "lower", tl.cex=0.5, tl.col="black", number.cex=0.3)

write.table(correlation_matrix,paste("C:\\Users\\dese0042\\Desktop\\layers\\correlation",".txt",sep=""),col.names = T,row.names = T,sep = "\t")


#### Removing unselected layers #### 
selected_rasters <- dropLayer(rasters, c('aspect_seg', 'aspect_std_seg', 'curv_seg', 'curv_std_seg', 'dist_rivers_seg', 'gndvi_seg', 'gndvi_std_seg', 'ndvi_seg', 'ndvi_std_seg','tpi_seg', 'tpi_std_seg','tri_seg','twi_std_seg'))

selected_training <-print(within(training, rm('aspect_seg', 'aspect_std_seg', 'curv_seg', 'curv_std_seg', 'dist_rivers_seg', 'gndvi_seg', 'gndvi_std_seg', 'ndvi_seg', 'ndvi_std_seg','tpi_seg', 'tpi_std_seg','tri_seg','twi_std_seg'))) #or selected_training <- training[,-(2:3)]


################################ Classification ################################
#### Train Artificial neural networks ####
set.seed(20)
fitControl<- trainControl(method = "repeatedcv", number = 5, repeats = 5)
ANN<- train(Class ~., data = selected_training, method = "nnet",preProcess = c("center", "scale"), trControl = fitControl)

timeStart <- proc.time() # measure computation time
Classification_ANN <-predict(selected_rasters,ANN)
proc.time() - timeStart # user time and system time.

writeRaster(Classification_ANN, "C:\\Users\\dese0042\\Desktop\\classification_ANN.tif")


#### Train Classification and regression trees ####
set.seed(20)
fitControl<- trainControl(method = "repeatedcv", number = 5, repeats = 5)
CART<-train(Class~.,data=selected_training, method="rpart1SE", trControl=fitControl)

timeStart <- proc.time() # measure computation time
Classification_CART <-predict(selected_rasters, CART)
proc.time() - timeStart # user time and system time.

writeRaster(Classification_CART, "C:\\Users\\dese0042\\Desktop\\classification_CART.tif")


#### Train K-Nearest neighbors ####
set.seed(20)
fitControl<- trainControl(method = "repeatedcv", number = 5, repeats = 5)
tuneGrid <- expand.grid(kmax = seq(11, 25, 2), 
                        distance = 2, 
                        kernel = c("triangular", "gaussian", "optimal"))
KNN<- train(Class ~., data=selected_training,  method = "kknn", tuneGrid = tuneGrid, trControl = fitControl)

timeStart <- proc.time() # measure computation time
Classification_KNN <-predict(selected_rasters,KNN)
proc.time() - timeStart # user time and system time.

writeRaster(Classification_KNN, "C:\\Users\\dese0042\\Desktop\\classification_KNN.tif")


#### Train Linear discriminant analysis ####
set.seed(20)
fitControl<- trainControl(method = "repeatedcv", number = 5, repeats = 5)
LDA<- train(Class ~., data = selected_training, method = "lda", trControl = fitControl)

timeStart <- proc.time() # measure computation time
Classification_LDA <-predict(selected_rasters, LDA)
proc.time() - timeStart # user time and system time.

writeRaster(Classification_LDA, "C:\\Users\\dese0042\\Desktop\\classification_LDA.tif")


#### Train Maximum likelihood ####
# We created the land cover map using ArcGIS Pro.


#### Train Naive bayes ####
set.seed(20) 
NB <- naiveBayes(Class ~ ., data = selected_training)

timeStart <- proc.time() # measure computation time
Classification_NB <-predict(selected_rasters,NB)
proc.time() - timeStart # user time and system time.

writeRaster(Classification_NB, "C:\\Users\\dese0042\\Desktop\\classification_NB.tif")


#### Train Random forests ####
set.seed(50)
RF <- randomForest(Class ~., data =selected_training ,type= "classification", ntree=1000, mtry = 3, importance=TRUE, na.action = na.roughfix)

timeStart <- proc.time() # measure computation time
Classification_RF <-predict(selected_rasters,RF)
proc.time() - timeStart # user time and system time.

writeRaster(Classification_RF, "C:\\Users\\dese0042\\Desktop\\classification_RF.tif")


#### Train Support vector machine ####
set.seed(20)
fitControl<- trainControl(method = "repeatedcv", number = 5, repeats = 5)
SVM<- train(Class ~., data = selected_training, method = "svmRadial", trControl = fitControl,  returnData= TRUE)

timeStart <- proc.time() # measure computation time
Classification_SVM <-predict(selected_rasters,SVM)
proc.time() - timeStart # user time and system time.

writeRaster(Classification_SVM, "C:\\Users\\dese0042\\Desktop\\classification_SVM.tif")


################################ Validation ################################
#### Confusion matrices ####
## Artificial neural networks ##
pred_ANN <- predict(ANN, newdata = validation)

matrix <- confusionMatrix(pred_ANN, validation$Class)
matrix

matrix2 <- as.matrix(matrix)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)


## Classification and regression trees ##
pred_CART <- predict(CART, newdata = validation)

matrix <- confusionMatrix(pred_CART, validation$Class)
matrix

matrix2 <- as.matrix(matrix)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)


## K-Nearest neighbors ##
pred_KNN <- predict(KNN, newdata = validation)

matrix <- confusionMatrix(pred_KNN, validation$Class)
matrix

matrix2 <- as.matrix(matrix)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)


## Linear discriminant analysis ##
pred_LDA <- predict(LDA, newdata = validation)

matrix <- confusionMatrix(pred_LDA, validation$Class)
matrix

matrix2 <- as.matrix(matrix)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)

## Maximum likelihood ##
ML <- raster(x = "C:\\Users\\dese0042\\Desktop\\ML\\ML_375_5.tif")

ML_extract<-as.data.frame(extract(ML, reference, na.rm=TRUE))
str(ML_extract)
names(ML_extract) <- c('Class')
ML_extract$Class[ML_extract$Class == '0'] <- 'bareground'
ML_extract$Class[ML_extract$Class == '1'] <- 'mesic'
ML_extract$Class[ML_extract$Class == '2'] <- 'snow'
ML_extract$Class[ML_extract$Class == '3'] <- 'water'
ML_extract$Class[ML_extract$Class == '4'] <- 'wetgrass'
ML_extract$Class[ML_extract$Class == '5'] <- 'wetmoss'
ML_extract$Class[ML_extract$Class == '6'] <- 'wetsedge'

Predict<-as.data.frame(ML_extract$Class)
names(Predict) <- c('Class_pred')

Testing<-as.data.frame(dataframe_ref$Class)
names(Testing) <- c('Class_ref')

combine<-cbind(Predict, Testing)
combine2 <- rapply(combine, as.factor, classes="character", how="replace")
str(combine2)

validation_ml <- as.data.frame(combine2[-intraining,])
str(validation_ml)
dim(validation_ml)

matrix<-confusionMatrix(validation_ml$Class_pred, validation_ml$Class_ref)

matrix2 <- as.matrix(matrix)
matrix2

rowSums(matrix2)
colSums(matrix2)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)


## Naive bayes ##
pred_NB <- predict(NB, newdata = validation)

matrix <- confusionMatrix(pred_NB, validation$Class)
matrix

matrix2 <- as.matrix(matrix)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)


## Random forests ##
pred_RF <- predict(RF, newdata = validation)

matrix <- confusionMatrix(pred_RF, validation$Class)
matrix 

matrix2 <- as.matrix(matrix)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)


## Support vector machine ##
pred_SVM <- predict(SVM, newdata = validation)

matrix <- confusionMatrix(pred_SVM, validation$Class)
matrix

matrix2 <- as.matrix(matrix)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)


## Ensemble classifier ##
EC <- raster(x = "C:\\Users\\dese0042\\Desktop\\Maj_voting\\Majority_voting\\result_map\\ensemble_classifier.tif")

EC_extract<-as.data.frame(extract(EC, reference, na.rm=TRUE))
str(EC_extract)
names(EC_extract) <- c('Class')
EC_extract$Class[EC_extract$Class == '1'] <- 'bareground'
EC_extract$Class[EC_extract$Class == '2'] <- 'mesic'
EC_extract$Class[EC_extract$Class == '3'] <- 'snow'
EC_extract$Class[EC_extract$Class == '4'] <- 'water'
EC_extract$Class[EC_extract$Class == '5'] <- 'wetgrass'
EC_extract$Class[EC_extract$Class == '6'] <- 'wetmoss'
EC_extract$Class[EC_extract$Class == '7'] <- 'wetsedge'

Predict<-as.data.frame(EC_extract$Class)
names(Predict) <- c('Class_pred')

Testing<-as.data.frame(dataframe_ref$Class)
names(Testing) <- c('Class_ref')

combine<-cbind(Predict, Testing)
combine2 <- rapply(combine, as.factor, classes="character", how="replace")
str(combine2)

validation_ec <- as.data.frame(combine2[-intraining,])
str(validation_ec)
dim(validation_ec)

matrix<-confusionMatrix(validation_ec$Class_pred, validation_ec$Class_ref)

matrix2 <- as.matrix(matrix)
matrix2

rowSums(matrix2)
colSums(matrix2)

# User accuracy
diag(matrix2) /rowSums(matrix2)

# Producer accuracy
diag(matrix2)/colSums(matrix2)

# Overall accuracy
o.accuracy <- sum(diag(matrix2))/sum(matrix2)
o.accuracy 

# Kappa
cohen.kappa(matrix2)
